//
// Simple fluid system for visualiztion purposes
//
// Author: Simon Green
// Email: sdkfeedback@nvidia.com
//
// Copyright (c) NVIDIA Corporation. All rights reserved.
////////////////////////////////////////////////////////////////////////////////

#include <math.h>

#include "FluidSystem.h"

FluidSystem::FluidSystem(CGcontext context, int width, int height, int depth)
  : m_cg_context(context),
    m_width(width), m_height(height), m_depth(depth),
    m_current(0)
{
    // we need 3 buffers - current, previous, and the state before that
    m_state[0] = new VolumeBuffer(GL_RGBA16F_ARB, width, height, depth, 1);
    m_state[1] = new VolumeBuffer(GL_RGBA16F_ARB, width, height, depth, 1);
    m_prevState = new VolumeBuffer(GL_RGBA16F_ARB, width, height, depth, 1);

    // load programs
    m_clear_prog = new Program(m_cg_context, "src/render_texture_3D/shaders/simulate.cg", "clear", CG_GL_FRAGMENT);
    m_copy_prog = new Program(m_cg_context, "src/render_texture_3D/shaders/simulate.cg", "copy", CG_GL_FRAGMENT);
    m_splat_prog = new Program(m_cg_context, "src/render_texture_3D/shaders/simulate.cg", "splat", CG_GL_FRAGMENT);
    m_step_prog = new Program(m_cg_context, "src/render_texture_3D/shaders/simulate.cg", "waves", CG_GL_FRAGMENT);

    m_step_prog->SetParameter("voxelSize", 1.0f / width, 1.0f / height, 1.0f / depth);
    m_step_prog->SetParameter("damping", 0.99f);

    reset();
}

FluidSystem::~FluidSystem()
{
    delete m_state[0];
    delete m_state[1];
    delete m_prevState;

    delete m_splat_prog;
    delete m_step_prog;
}

inline float
FluidSystem::frand()
{
    return rand() / (float) RAND_MAX;
}

// add a random splat to the simulation
void
FluidSystem::splat()
{
    m_splat_prog->SetParameter("centre", frand(), frand(), frand());
    m_splat_prog->SetParameter("radius", 0.1f);
    m_splat_prog->SetParameter("color", frand(), frand(), frand(), 1.0f);

    m_state[m_current]->setBlendMode(VolumeBuffer::BLEND_ADDITIVE);
    m_state[m_current]->runProgram(m_splat_prog);
    m_state[m_current]->setBlendMode(VolumeBuffer::BLEND_NONE);
}

// reset the simulation by clearing buffers
void
FluidSystem::reset()
{
    m_clear_prog->SetParameter("color", 0.0, 0.0, 0.0, 0.0);
    m_state[m_current]->runProgram(m_clear_prog);
    m_state[1 - m_current]->runProgram(m_clear_prog);
    m_prevState->runProgram(m_clear_prog, 0);
}

// step the simulation
void
FluidSystem::step(float timestep)
{
    // copy previous state
    m_copy_prog->SetTexture("tex", m_state[1 - m_current]->getTexture());
    m_prevState->runProgram(m_copy_prog);   

    // compute new state
    m_step_prog->SetTexture("currentTex", m_state[m_current]->getTexture());
    m_step_prog->SetTexture("prevTex", m_prevState->getTexture());
    m_state[1 - m_current]->runProgram(m_step_prog);

    m_current = 1 - m_current;

}
