//
//  This sample demonstrates rendering to a 3D texture using the framebuffer object extension.
//  This feature is only supported on GeForce 8 series GPUs.
//
//  A simple wave simulation is run in a 3D volume.
//  The 3D texture is then displayed using ray marching.
//
// Author: Simon Green
// Email: sdkfeedback@nvidia.com
//
// Copyright (c) NVIDIA Corporation. All rights reserved.
////////////////////////////////////////////////////////////////////////////////

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <map>

#include <GL/glew.h>
#include <GL/glut.h>
#include <Cg/cg.h>

#include "nvGlutManipulators.h"
#include "FluidSystem.h"
#include "VolumeRender.h"

using std::map;

enum UIOption {
    OPTION_DRAW_CUBE,
    OPTION_SIMULATE,
    OPTION_AUTO_SEED,
    OPTION_COUNT
};
bool options[OPTION_COUNT];
map<char, UIOption> optionKeyMap;

int width = 800;
int height = 600;

nv::GlutExamine trackball;

CGcontext cgContext; 
FluidSystem *fluidSys;
VolumeRender *volumeRender;

bool autoSeed = true;

void display()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    trackball.applyTransform();

    // run simulation
    if (options[OPTION_SIMULATE]) {
        if ( options[OPTION_AUTO_SEED] && ((rand() & 0xff) > 220)) {
            fluidSys->splat();
        }
        fluidSys->getStateBuffer()->setFiltering(GL_NEAREST);
        fluidSys->step(1.0f);
    }

    // render volume
    glViewport(0, 0, width, height);
    fluidSys->getStateBuffer()->setFiltering(GL_LINEAR);
    volumeRender->setVolume(fluidSys->getStateBuffer());
    volumeRender->render();

    if (options[OPTION_DRAW_CUBE]) {
        glutWireCube(2.0f);
    }

    glutSwapBuffers();
}

void mouse( int button, int state, int x, int y)
{
    trackball.mouse(button, state, x, y);
    glutPostRedisplay();
}

void motion(int x, int y)
{
    trackball.motion(x, y);
    glutPostRedisplay();
}

void idle()
{
    trackball.idle();
    glutPostRedisplay();
}

void key(unsigned char c, int x, int y)
{
    c = tolower(c);

    if (optionKeyMap.find(c) != optionKeyMap.end())
        options[optionKeyMap[c]] = ! options[optionKeyMap[c]];

    switch (c) {
        case '\033': //escape key
        case 'q':
            exit(0);
            break;

        case 13:
            fluidSys->step(1.0f);
            break;

        case 'r':
            fluidSys->reset();
            break;

        case 'x':
            fluidSys->splat();
            break;
    }

    glutPostRedisplay();
}

void reshape(int w, int h)
{
    width = w;
    height = h;

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(40.0, (double)width / height, 0.1, 10.0);
    
    glViewport(0, 0, width, height);

    trackball.reshape(w, h);
}

void mainMenu(int i)
{
    key((unsigned char) i, 0, 0);
}

void initMenus() {
    glutCreateMenu(mainMenu);
    glutAddMenuEntry("Toggle simulation [ ]", ' ');
    glutAddMenuEntry("Step simulation [ret]", 13);
    glutAddMenuEntry("Toggle auto seeding [a]", 'a');
    glutAddMenuEntry("Add splat [x]", 'x');
    glutAddMenuEntry("Toggle cube [c]", 'c');
    glutAddMenuEntry("Quit [esc]", '\033');
    glutAttachMenu(GLUT_RIGHT_BUTTON);
}

void initGL()
{
    glewInit();
    if (!glewIsSupported( "GL_VERSION_2_0 GL_EXT_framebuffer_object" 
        )) {
        printf( "Required extensions not supported.\n");
        exit(-1);
    }

    glEnable(GL_DEPTH_TEST);
}

void cgErrorCallback()
{
    CGerror lastError = cgGetError();
    if(lastError)
    {
        printf("%s\n", cgGetErrorString(lastError));
        printf("%s\n", cgGetLastListing(cgContext));
        exit(1);
    }
}

int main(int argc, char** argv)
{
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
    glutInitWindowSize(width, height);
    glutCreateWindow("render_texture_3D");

    glutDisplayFunc(display);
    glutKeyboardFunc(key);
    glutReshapeFunc(reshape);
    glutMouseFunc(mouse);
    glutMotionFunc(motion);
    glutIdleFunc(idle);

    initGL();
    initMenus();

    trackball.setDollyActivate(GLUT_LEFT_BUTTON, GLUT_ACTIVE_CTRL);
    trackball.setPanActivate(GLUT_LEFT_BUTTON, GLUT_ACTIVE_SHIFT);
    trackball.setDollyPosition(-5.0f);

    // create simulation
    cgContext = cgCreateContext();
    cgSetErrorCallback(cgErrorCallback);

    fluidSys = new FluidSystem(cgContext, 128, 128, 128);
    fluidSys->splat();

    volumeRender = new VolumeRender(cgContext , fluidSys->getStateBuffer());
    volumeRender->setDensity(0.5);
    volumeRender->setBrightness(1.0);

    //setup the option keys
    optionKeyMap['c'] = OPTION_DRAW_CUBE;
    options[OPTION_DRAW_CUBE] = false;

    optionKeyMap[' '] = OPTION_SIMULATE;
    options[OPTION_SIMULATE] = true;

    optionKeyMap['a'] = OPTION_AUTO_SEED;
    options[OPTION_AUTO_SEED] = true;

    printf( "render_texture_3D - sample demonstrating rendering to a 3D texture\n");
    printf( "   q/[ESC]    - Quit the app\n");
    printf( "   [SPACE]    - Toggle continuous wave simulation\n");
    printf( "      a       - Toggle automatically seeding the volume with splats\n");
    printf( "      r       - Reset the simulation\n");
    printf( "      x       - Add a splat to the simulation\n");
    printf( "      c       - Toggle drawing the extents of the volume in wireframe\n");
    printf( "   [Enter]    - Step the simulation forward one cycle\n");
    printf( "   [MOUSE]    - Left - rotate, Left + Ctrl - dolly, Left + Shift - pan\n");
    printf( "   [MOUSE]    - Right - application menu\n\n");

    glutMainLoop();

    return 0;
}
